/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Generic FX class
 * Leto/calodox, 28.12.2000
 * Rewrite: leto, 16.3.2001
 */

/*! 	\file fx.h
	\brief Effect interface with demoniac.
	\author Julien Pilet (leto@calodox.org)
	\version $Revision: 1.4 $

	Demoniac provides to effects a good option managment and interpolation 
	system.
 */

#ifndef _FX_H_
#define _FX_H_

#include <math.h>
#include <limits.h>
#include <float.h>

// autoconf integration
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*!
 * possible flags for an option 
 */
enum { 
	//! init() is called each time the option is modified
	OPTION_REINIT = 1,								
	//! option can't be modified
	OPTION_READ_ONLY = 2,		

	// for bounding values.
	OPTION_LIMIT_CLAMP = 0,		//!< clamp to min/max (this is default)
	OPTION_LIMIT_WRAP = 4		//!< wrap to min (max) when reaching max (min)
};

/*
 * Color access function.
 */

#ifdef WORDS_BIGENDIAN
        typedef struct {
                unsigned char a,b,g,r;
        } t_rgba;
#else   
        typedef struct {
                unsigned char r,g,b,a;
        } t_rgba;
#endif

typedef union {
	unsigned char comp[4];
        t_rgba  rgb;
        unsigned int i;		// unsigned int should be 32 bits.
} t_color;

/*!
 * class for framebuffer emulation
 */

class DrawBuffer {
public:
	const char *name;	//!< name of the buffer, 0 if none.
	int x, y;	//!< size of the buffer, in pixels
	t_color *buff;	//!< the buffer itself. 0 if not allocated.
	void *data;	//!< free for effects. 0 if unused.

	DrawBuffer() : name(0), x(0), y(0), buff(0), data(0) 
	{
	}

	~DrawBuffer() {
		if (buff) delete buff;
		//if (name) delete name;
	}
};
	
/*! 
  Register a floating point option.
  \param optName 	name of this option
  \param flags 		see enum before
  \param value  	a pointer to the float, initialized to a default value.
  \param min		(*value) will be bounded to [min, max]
 */		
void registerFloatOpt(char *optName, int flags, float *value, 
	float min = -FLT_MAX, float max= FLT_MAX);

void registerIntOpt(char *optName, int flags, int *value,
		int min = INT_MIN, int max = INT_MAX);

//! takes an array of *4* floats (R,G,B,A)
void registerColorOpt(char *name, int flags, float *values);

//! a boolean option.
void registerBoolOpt(char *optName, int flags, bool *value);

//! a 3D position
void register3DPosOpt(char *optName, int flags, float *values);

//! a 3D rotation matrix (homogeneous)
void register3DRotOpt(char *optName, int flags, float value[4][4]);

/*! a string option. Actually, an array of maxLen bytes ready to
 receive a string of maxLen - 1 characters. */
void registerStringOpt(char *optName, int flags, char *string, int maxLen);

/*! \brief register a DrawBuffer. 
   Read + write is possible. Don't bother allocating
   or freeing memory. Don't assume anything about buffer size.
*/
void registerDrawBufferOpt(char *optName, int flags, DrawBuffer **buffer);

/*! \brief register a screen position. 
  values points on a 2d array.  values are between 0 and 1, from top to down,
  left to right.
*/
void registerScreenPosOpt(char *optName, int flags, float *values);

/*! \brief rectangle on screen. 
 \param values values are between 0 and 1 
 values[0] = left;  
 values[1] = right; 
 values[2] = top; 
 values[3] = bottom
*/
void registerRectangleOpt(char *optName, int flags, float *values);

/*********************************
**     Generic effect class     **
*********************************/

//! Generic effect class
class FXGeneric {
public:

	/*!
	 * virtual constructor. Basically 
	 * create a new object of the same class.
	 */
	virtual FXGeneric *create()=0;

	/*!
	 * returns the effect name, a small 
	 * description, or 0.
	 */
	virtual char *getName(void)=0;

	/*!
	 * In this function, the effect should call register*Opt
	 * functions, or do nothing, if it has no options.
	 */
	virtual void registerOptions()=0;

	/*!
	 * get the effect ready to render.
	 * returns 0 if ok, any non-null error code 
	 * if there is a problem.
	 */
	virtual int init(void)=0;

	/*!
	 * run the effect in the current openGL context.
	 * returns 0 if ok, any non-zero error code if error.
	 */
	virtual int render()=0;

	/*!
	 * the fx should free all used memory.
	 */
	virtual void finish(void)=0;

	virtual ~FXGeneric(){} 
};

#endif // _GENERIC_FX_H_
