/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Time line widget for Gtk.
 * leto/calodox, 7 2001
 * (leto@calodox.org)
 */
#include <math.h>
#include <stdio.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "gtktimeline.h"

#define SCROLL_DELAY_LENGTH  300
#define TIME_LINE_DEFAULT_WIDTH 100
#define TIME_LINE_DEFAULT_HEIGHT 28
#define GTK_TIME_LINE_KEY_WIDTH (GTK_WIDGET(widget)->allocation.height / 4)

/* signals enum 
 * Note that key means time key, not keyboard key. 
 */
enum {
	KEY_PRESS,
	LAST_SIGNAL
};

/* Forward declarations */

static void gtk_time_line_class_init               (GtkTimeLineClass    *klass);
static void gtk_time_line_init                     (GtkTimeLine         *time_line);
static void gtk_time_line_destroy                  (GtkObject        *object);
static void gtk_time_line_realize                  (GtkWidget        *widget);
static void gtk_time_line_size_request             (GtkWidget      *widget,
					       GtkRequisition *requisition);
static void gtk_time_line_size_allocate            (GtkWidget     *widget,
					       GtkAllocation *allocation);
static gint gtk_time_line_expose                   (GtkWidget        *widget,
						GdkEventExpose   *event);
static gint gtk_time_line_button_press             (GtkWidget        *widget,
						GdkEventButton   *event);
static gint gtk_time_line_button_release           (GtkWidget        *widget,
						GdkEventButton   *event);
static gint gtk_time_line_motion_notify            (GtkWidget        *widget,
						GdkEventMotion   *event);
static gint gtk_time_line_timer                    (GtkTimeLine         *time_line);

static void gtk_time_line_update_mouse             (GtkTimeLine *time_line, gint x, gint y);
static void gtk_time_line_update                   (GtkTimeLine *time_line);
static void gtk_time_line_adjustment_changed       (GtkAdjustment    *adjustment,
						gpointer          data);
static void gtk_time_line_adjustment_value_changed (GtkAdjustment    *adjustment,
						gpointer          data);
static void gtk_time_line_key_value_changed (GtkAdjustment    *adjustment,
						gpointer          data);
static gint gtk_time_line_key_compare(gconstpointer a, gconstpointer b);
static gint gtk_time_line_get_xpixel_from_value(GtkTimeLine *time_line, GtkAdjustment *adjustment);
static gfloat gtk_time_line_get_value_from_xpixel(GtkTimeLine *time_line, int x, int xc);

static void gtk_time_line_real_key_press(GtkTimeLine *timeline, GdkEventButton *button, GtkAdjustment *key);

static gint gtk_time_line_signals[LAST_SIGNAL] = { 0 };

/* Local data */

static GtkWidgetClass *parent_class = NULL;

guint
gtk_time_line_get_type ()
{
	static guint time_line_type = 0;

	if (!time_line_type)
	{
		GtkTypeInfo time_line_info =
		{
			"GtkTimeLine",
			sizeof (GtkTimeLine),
			sizeof (GtkTimeLineClass),
			(GtkClassInitFunc) gtk_time_line_class_init,
			(GtkObjectInitFunc) gtk_time_line_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL,
		};

		time_line_type = gtk_type_unique (gtk_widget_get_type (), &time_line_info);
	}

	return time_line_type;
}

static void
gtk_time_line_class_init (GtkTimeLineClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = (GtkObjectClass*) klass;
	widget_class = (GtkWidgetClass*) klass;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy = gtk_time_line_destroy;

	widget_class->realize = gtk_time_line_realize;
	widget_class->expose_event = gtk_time_line_expose;
	widget_class->size_request = gtk_time_line_size_request;
	widget_class->size_allocate = gtk_time_line_size_allocate;
	widget_class->button_press_event = gtk_time_line_button_press;
	widget_class->button_release_event = gtk_time_line_button_release;
	widget_class->motion_notify_event = gtk_time_line_motion_notify;

	klass->key_press = gtk_time_line_real_key_press;

	gtk_time_line_signals[KEY_PRESS] =
		gtk_signal_new( "key_press",
				GTK_RUN_LAST | GTK_RUN_ACTION, 
				object_class->type,
				GTK_SIGNAL_OFFSET(GtkTimeLineClass, key_press),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2,
				GTK_TYPE_POINTER,
				GTK_TYPE_POINTER);

	gtk_object_class_add_signals(object_class, gtk_time_line_signals, LAST_SIGNAL);
}

/* GtkTimeLine constructor */
static void
gtk_time_line_init (GtkTimeLine *time_line)
{
  time_line->button = 0;
  time_line->policy = GTK_UPDATE_CONTINUOUS;
  time_line->timer = 0;
  time_line->adjustment = NULL;
  time_line->key_list = 0;
  time_line->key_grab = 0;
  time_line->lower = -G_MAXFLOAT;
  time_line->upper = G_MAXFLOAT;
}

GtkWidget*
gtk_time_line_new (GtkAdjustment *adjustment)
{
  GtkTimeLine *time_line;

  time_line = gtk_type_new (gtk_time_line_get_type ());

  if (!adjustment)
    adjustment = (GtkAdjustment*) gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

  gtk_time_line_set_adjustment (time_line, adjustment);

  return GTK_WIDGET (time_line);
}

static void
gtk_time_line_destroy (GtkObject *object)
{
  GtkTimeLine *time_line;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (object));

  time_line = GTK_TIME_LINE (object);

  if (time_line->adjustment) {
      gtk_signal_disconnect_by_data (GTK_OBJECT (time_line->adjustment), (gpointer) time_line);
      gtk_object_unref (GTK_OBJECT (time_line->adjustment));
  }

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

GtkAdjustment*
gtk_time_line_get_adjustment (GtkTimeLine *time_line)
{
  g_return_val_if_fail (time_line != NULL, NULL);
  g_return_val_if_fail (GTK_IS_TIME_LINE (time_line), NULL);

  return time_line->adjustment;
}

void
gtk_time_line_set_update_policy (GtkTimeLine      *time_line,
			     GtkUpdateType  policy)
{
  g_return_if_fail (time_line != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (time_line));

  time_line->policy = policy;
}

void
gtk_time_line_set_adjustment (GtkTimeLine      *time_line,
			  GtkAdjustment *adjustment)
{
  g_return_if_fail (time_line != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (time_line));

  if (adjustment == time_line->adjustment)
	  return;

  if (time_line->adjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (time_line->adjustment), (gpointer) time_line);
      gtk_object_unref (GTK_OBJECT (time_line->adjustment));
    }

  time_line->adjustment = adjustment;
  gtk_object_ref (GTK_OBJECT (time_line->adjustment));

  gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
		      (GtkSignalFunc) gtk_time_line_adjustment_changed,
		      (gpointer) time_line);
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
		      (GtkSignalFunc) gtk_time_line_adjustment_value_changed,
		      (gpointer) time_line);

  time_line->old_value = adjustment->value;
  time_line->old_lower = adjustment->lower;
  time_line->old_upper = adjustment->upper;

  gtk_time_line_update (time_line);
}

void
gtk_time_line_set_range (GtkTimeLine *time_line, gfloat lower, gfloat upper) {

  g_return_if_fail (time_line != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (time_line));
  g_return_if_fail(lower < upper);

  time_line->lower = lower;
  time_line->upper = upper;

  gtk_widget_queue_draw (GTK_WIDGET (time_line));
}


static void
gtk_time_line_realize (GtkWidget *widget)
{
  GtkTimeLine *time_line;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  time_line = GTK_TIME_LINE (widget);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events (widget) | 
    GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
    GDK_POINTER_MOTION_HINT_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_user_data (widget->window, widget);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);
}

static void 
gtk_time_line_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  requisition->width = TIME_LINE_DEFAULT_WIDTH;
  requisition->height = TIME_LINE_DEFAULT_HEIGHT;
}

static void
gtk_time_line_size_allocate (GtkWidget     *widget,
			GtkAllocation *allocation)
{
  GtkTimeLine *time_line;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;
  time_line = GTK_TIME_LINE (widget);

  if (GTK_WIDGET_REALIZED (widget))
    {

      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

    }
}

gint gtk_time_line_get_xpixel_from_value(GtkTimeLine *time_line, GtkAdjustment *adjustment) {
	gfloat left, right;
	gdouble percent;

	if (time_line->lower > time_line->adjustment->lower)
		left = time_line->lower;
	else 
		left = time_line->adjustment->lower;

	if (time_line->upper < time_line->adjustment->upper)
		right = time_line->upper;
	else
		right = time_line->adjustment->upper;

	  percent = (adjustment->value - left) / (right - left);
	  return ((gint) (percent * (gdouble) (GTK_WIDGET(time_line)->allocation.width-3))) + 2;
}

static gfloat 
gtk_time_line_get_value_from_xpixel(GtkTimeLine *time_line, int x, int xc) {
	gfloat left, right;
	gfloat percent;

	if (time_line->lower > time_line->adjustment->lower)
		left = time_line->lower;
	else 
		left = time_line->adjustment->lower;

	if (time_line->upper < time_line->adjustment->upper)
		right = time_line->upper;
	else
		right = time_line->adjustment->upper;

	percent = ((gfloat) (x-2))/(gfloat)(xc-3);
	return percent * (right - left) + left;
}



static gint
gtk_time_line_expose (GtkWidget      *widget,
		 GdkEventExpose *event)
{
  GtkTimeLine *time_line;
  GSList *gslist;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_TIME_LINE (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->count > 0)
    return FALSE;
  
  time_line = GTK_TIME_LINE (widget);

  gtk_paint_box(widget->style, widget->window,
		  GTK_WIDGET_STATE(widget),
		  GTK_SHADOW_IN,
		  &event->area, widget, "timeline",
		  0, 0,
		  widget->allocation.width,
		  widget->allocation.height);

  gtk_paint_hline(widget->style, widget->window,
		  GTK_WIDGET_STATE(widget),
		  &event->area, widget, "timeline",
		  2,
		  widget->allocation.width -3,
		  widget->allocation.height/2);
		  
  if (time_line->adjustment) {
	  gint posx = gtk_time_line_get_xpixel_from_value(time_line, time_line->adjustment);

	  gdk_draw_line (widget->window,
			  widget->style->fg_gc[widget->state],
			  posx,
			  2,
			  posx,
			  widget->allocation.height-2);

	  /* Draw all keys */
	  for (gslist = time_line->key_list; gslist != 0; gslist = gslist->next) {
		  GdkPoint points[3];
		  int pos_up, pos_down;
		  GtkAdjustment *adj = (GtkAdjustment *) gslist->data;

		  if (adj->value < time_line->lower || adj->value > time_line->upper)
			  continue;

		  posx = gtk_time_line_get_xpixel_from_value(time_line,
				  (GtkAdjustment *) gslist->data);

		  pos_up = widget->allocation.height / 8;
		  pos_down = (7*widget->allocation.height)/8;

		  points[0].x = posx - (GTK_TIME_LINE_KEY_WIDTH/2);
		  points[0].y = pos_down;
		  points[1].x = posx + (GTK_TIME_LINE_KEY_WIDTH/2);
		  points[1].y = pos_down;
		  points[2].x = posx;
		  points[2].y = widget->allocation.height / 2;

		  /*
		  gdk_draw_polygon(widget->window,
			  widget->style->fg_gc[widget->state],
			  TRUE, // filled
			  points,
			  3);
		  */
		  gtk_paint_polygon(widget->style, widget->window,
				  GTK_STATE_PRELIGHT,
				  GTK_SHADOW_OUT,
				  &event->area, widget, "timeline-key",
				  points,
				  3,
				  TRUE);

		  gdk_draw_line(widget->window,
			  widget->style->fg_gc[widget->state],
			  posx,
			  widget->allocation.height / 2,
			  posx,
			  pos_up);
	  }
  }

  return FALSE;
}

static gint
gtk_time_line_button_press (GtkWidget      *widget,
		       GdkEventButton *event)
{
	GtkTimeLine *time_line;
	GSList *gslist;
	gint posx;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TIME_LINE (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	time_line = GTK_TIME_LINE (widget);

	gtk_grab_add (widget);

	time_line->button = event->button;


	/* did the user clic on a key ? */
	if (event->y >= (widget->allocation.height /2)) {
		for (gslist = time_line->key_list; gslist != 0; gslist = gslist->next) {
			posx = gtk_time_line_get_xpixel_from_value(time_line,
					(GtkAdjustment *) gslist->data);

			if (event->x >= posx - (GTK_TIME_LINE_KEY_WIDTH/2) &&
				event->x <= posx + (GTK_TIME_LINE_KEY_WIDTH/2)) {

				if (time_line->button == 3) {
					gtk_signal_emit(GTK_OBJECT(widget),
							gtk_time_line_signals[KEY_PRESS], 
							event, gslist->data);
				} else {
					time_line->key_grab = (GtkAdjustment *) gslist->data;
				}
			}
		}
	}

	gtk_time_line_update_mouse (time_line, event->x, event->y);

	return TRUE;
}

static gint
gtk_time_line_button_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
	GtkTimeLine *time_line;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TIME_LINE (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	time_line = GTK_TIME_LINE (widget);

	if (time_line->button == event->button)
	{
		gtk_grab_remove (widget);

		time_line->button = 0;

		if (time_line->policy == GTK_UPDATE_DELAYED)
			gtk_timeout_remove (time_line->timer);

		if (time_line->policy != GTK_UPDATE_CONTINUOUS) { 
			if (time_line->old_value != time_line->adjustment->value) 
				gtk_signal_emit_by_name (GTK_OBJECT (time_line->adjustment), "value_changed");

			if (time_line->key_grab && (time_line->key_grab_old_value != time_line->key_grab->value)) {
				gtk_signal_emit_by_name(GTK_OBJECT (time_line->adjustment), 
						"value_changed");
			}
		}

		if (time_line->key_grab) {
			time_line->key_grab=0;
		}
	}

	return FALSE;
}

static gint
gtk_time_line_motion_notify (GtkWidget      *widget,
			 GdkEventMotion *event)
{
  GtkTimeLine *time_line;
  GdkModifierType mods;
  gint x, y, mask;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_TIME_LINE (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  time_line = GTK_TIME_LINE (widget);

  if (time_line->button != 0)
    {
      x = event->x;
      y = event->y;

      if (event->is_hint || (event->window != widget->window))
	gdk_window_get_pointer (widget->window, &x, &y, &mods);

      switch (time_line->button)
	{
	case 1:
	  mask = GDK_BUTTON1_MASK;
	  break;
	case 2:
	  mask = GDK_BUTTON2_MASK;
	  break;
	case 3:
	  mask = GDK_BUTTON3_MASK;
	  break;
	default:
	  mask = 0;
	  break;
	}

      if (mods & mask)
	gtk_time_line_update_mouse (time_line, x,y);
    }

  return FALSE;
}

static gint
gtk_time_line_timer (GtkTimeLine *time_line)
{
	g_return_val_if_fail (time_line != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TIME_LINE (time_line), FALSE);

	if (time_line->policy == GTK_UPDATE_DELAYED) {
		gtk_signal_emit_by_name (GTK_OBJECT (time_line->adjustment), "value_changed");
		if (time_line->key_grab) 
			gtk_signal_emit_by_name(GTK_OBJECT (time_line->key_grab), "value_changed");
	}

	return FALSE;
}


static void
gtk_time_line_update_mouse (GtkTimeLine *widget, gint x, gint y)
{
  gint xc, yc;
  gfloat old_value, key_old_value;
  GSList *gslist;
  gint posx;
  gint magnet = FALSE;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (widget));

  xc = GTK_WIDGET(widget)->allocation.width;
  yc = GTK_WIDGET(widget)->allocation.height / 2;

  old_value = widget->adjustment->value;

  /* search for a key to magnet to */
  for (gslist=widget->key_list; gslist !=0; gslist = gslist->next) {
	  posx = gtk_time_line_get_xpixel_from_value(widget,
			  (GtkAdjustment *) gslist->data);

	  if ((GtkAdjustment *) gslist->data != widget->key_grab &&
			  x >= posx - (GTK_TIME_LINE_KEY_WIDTH/2) &&
			  x <= posx + (GTK_TIME_LINE_KEY_WIDTH/2)) {
		  magnet = TRUE;
		  widget->adjustment->value = ((GtkAdjustment *)gslist->data)->value;
		  break;
	  }
  }
	
  if (!magnet) {
	  // let's compute the time from the clic coordinates

	  widget->adjustment->value = 
		  gtk_time_line_get_value_from_xpixel(widget, x, xc);
  }

  if (widget->key_grab) {
	  key_old_value = widget->key_grab->value;
	  if (widget->adjustment->value < widget->key_grab->lower)
		  widget->key_grab->value = widget->key_grab->lower;
	  else {
		  if (widget->adjustment->value > widget->key_grab->upper)
			  widget->key_grab->value = widget->key_grab->upper;
		  else
			  widget->key_grab->value = widget->adjustment->value;
	  }	
  }

  if (widget->adjustment->value != old_value)
    {
      if (widget->policy == GTK_UPDATE_CONTINUOUS)
	{
	  if (widget->key_grab && key_old_value != widget->key_grab->value)
		  gtk_signal_emit_by_name (GTK_OBJECT (widget->key_grab), "value_changed");
	  gtk_signal_emit_by_name (GTK_OBJECT (widget->adjustment), "value_changed");
		  
	}
      else
	{
	  gtk_widget_queue_draw (GTK_WIDGET(widget));

	  if (widget->policy == GTK_UPDATE_DELAYED)
	    {
	      if (widget->timer)
		gtk_timeout_remove (widget->timer);

	      widget->timer = gtk_timeout_add (SCROLL_DELAY_LENGTH,
					     (GtkFunction) gtk_time_line_timer,
					     (gpointer) widget);
	    }
	}
    }
}

static void
gtk_time_line_update (GtkTimeLine *time_line)
{
  gfloat new_value;
  
  g_return_if_fail (time_line != NULL);
  g_return_if_fail (GTK_IS_TIME_LINE (time_line));

  new_value = time_line->adjustment->value;
  
  if (new_value < time_line->adjustment->lower)
    new_value = time_line->adjustment->lower;

  if (new_value > time_line->adjustment->upper)
    new_value = time_line->adjustment->upper;

  if (new_value != time_line->adjustment->value)
    {
      time_line->adjustment->value = new_value;
      gtk_signal_emit_by_name (GTK_OBJECT (time_line->adjustment), "value_changed");
    }

  //gtk_widget_draw (GTK_WIDGET(time_line), NULL);
  gtk_widget_queue_draw(GTK_WIDGET(time_line)); 
}

static void
gtk_time_line_adjustment_changed (GtkAdjustment *adjustment,
			      gpointer       data)
{
  GtkTimeLine *time_line;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);

  time_line = GTK_TIME_LINE (data);

  if ((time_line->old_value != adjustment->value) ||
      (time_line->old_lower != adjustment->lower) ||
      (time_line->old_upper != adjustment->upper))
    {
      gtk_time_line_update (time_line);

      time_line->old_value = adjustment->value;
      time_line->old_lower = adjustment->lower;
      time_line->old_upper = adjustment->upper;
    }
}

static void
gtk_time_line_adjustment_value_changed (GtkAdjustment *adjustment,
				    gpointer       data)
{
	GtkTimeLine *time_line;

	g_return_if_fail (adjustment != NULL);
	g_return_if_fail (data != NULL);

	time_line = GTK_TIME_LINE (data);

	if (time_line->old_value != adjustment->value)
	{
		gtk_time_line_update (time_line);

		time_line->old_value = adjustment->value;
	}
}

void gtk_time_line_key_value_changed (GtkAdjustment    *adjustment,
		gpointer data) {
	GtkTimeLine *time_line = data;
	gfloat new_value;

	g_return_if_fail (adjustment != NULL);
	g_return_if_fail (GTK_IS_ADJUSTMENT(adjustment));

	time_line = GTK_TIME_LINE (data);

	new_value = adjustment->value;

	if (time_line->key_grab_old_value == new_value)
		return;

	if (new_value < adjustment->lower)
		new_value = adjustment->lower;

	if (new_value > adjustment->upper)
		new_value = adjustment->upper;

	if (new_value != adjustment->value)
	{
		adjustment->value = new_value;
		gtk_signal_emit_by_name (GTK_OBJECT (time_line->adjustment), "value_changed");
	}

	gtk_widget_queue_draw (GTK_WIDGET(time_line));

	time_line->key_grab_old_value = new_value;
}

gint gtk_time_line_key_compare(gconstpointer a, gconstpointer b) {

	GtkAdjustment *adjA = (GtkAdjustment *) a;
	GtkAdjustment *adjB = (GtkAdjustment *) b;

	return (adjA->value <= adjB->value);
}


void gtk_time_line_add_key(GtkTimeLine *time_line, GtkAdjustment *key) {

	g_return_if_fail (time_line != NULL);
	g_return_if_fail (GTK_IS_TIME_LINE (time_line));

	time_line->key_list = g_slist_insert_sorted(time_line->key_list, 
			key, gtk_time_line_key_compare);

	gtk_object_ref(GTK_OBJECT(key));

	gtk_signal_connect(GTK_OBJECT(key), "value_changed", 
			GTK_SIGNAL_FUNC(gtk_time_line_key_value_changed), time_line);

	if (GTK_WIDGET_VISIBLE(GTK_WIDGET(time_line)))
			gtk_widget_queue_draw(GTK_WIDGET(time_line));

}
	
gint gtk_time_line_remove_key(GtkTimeLine *time_line, GtkAdjustment *key) {

	g_return_val_if_fail (time_line != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TIME_LINE (time_line), FALSE);

	if (g_slist_find(time_line->key_list, key)) {

		time_line->key_list = g_slist_remove(time_line->key_list, key);
		gtk_object_unref(GTK_OBJECT(key));

		if (GTK_WIDGET_VISIBLE(GTK_WIDGET(time_line)))
			gtk_widget_queue_draw(GTK_WIDGET(time_line));

		return TRUE;
	}
	return FALSE;
}

static void 
gtk_time_line_real_key_press(GtkTimeLine *timeline, GdkEventButton *button, GtkAdjustment *key) {
	g_return_if_fail (timeline != NULL);
	g_return_if_fail (GTK_IS_TIME_LINE (timeline));

	/* there is actually nothing to do. It might change in the future */
}

