/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * loader.cpp: demoniac file loader
 * leto/calodox, 3.2001
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "loader.h"

#define warning(a) fprintf(stderr, "%s(%d): %s", filename, line, a)

static char *typeNames[] = { "NONE", "OBJECT", "OBJECT_END", "STRING", "FLOAT", "INT", "FLOAT_VECT", "INT_VECT" };
Loader::Loader() {
	in = 0;
	name[0]=0;
	line=0;
	filename="";
	type = NONE;
}

Loader::~Loader() {
	delete[] filename;
	if (in)
		fclose(in);
}

void Loader::error(char *string) {
	fprintf(stderr, "%s(line: %d. Name: %s. Type: %s. Parsing: %s):\n\t%s\n",
		filename,
		line,
		name,
		typeNames[type],
		buffer,
		string);
	assert(0);
}

bool Loader::loadFile(char *fileName) {
	if (!fileName) return false;
	
	in = fopen(fileName, "rt");
	line = 0;
	this->filename = new char[strlen(fileName)+1];
	strcpy(this->filename, fileName);
	
	if (in) {
		nextLine();
		return true;
	} else {
		return false;
	}
}

void Loader::nextToken() {
	const char *blanks=" \t\r\n";
	const char *endstring = "\"";
	const char *stop;
	int c;
	int count = 0;

	// eat all blanks
	do { 
		c= fgetc(in);
		if (c=='\n') line++;
	} while (strchr(blanks,c));

	// is this a string ?
	if (c=='"') {
		stop = endstring;
	} else {
		stop = blanks;
		buffer[count++] = c;
	}

	while (count < (LOADER_BUFFER_LENGTH - 1)) {
		c = fgetc(in);
		if (c == '\n') line++;

		if (strchr(stop, c) || c == EOF) {
			buffer[count]=0;
			break;
		} else {
			buffer[count++]=c;
		}
	} 

	buffer[LOADER_BUFFER_LENGTH-1]=0;
}

bool Loader::nextLine() {
	if (!in) return false;

	// read type
	nextToken();

	if (strcmp("object", buffer)==0) {
		type = OBJECT;
	} else 
	if (strcmp("}", buffer)==0) {
		type = OBJECT_END;
	} else 
	if (strcmp("string", buffer)==0) {
		type = STRING;
	} else 
	if (strcmp("float", buffer) ==0) {
		type = FLOAT;
	} else 
	if (strcmp("int", buffer)==0) {
		type = INT;
	} else 
	if (strcmp("fvect", buffer)==0) {
		type=FLOAT_VECT;
	} else 
	if (strcmp("ivect", buffer) == 0) {
		type = INT_VECT;
	} else {
		// what is it then ?
		error("buffer: unknown type");
		return false;
	}

	// let's get the name
	if (type != OBJECT_END) {
		nextToken();
		strncpy(name, buffer, 127);
		name[127]=0;
	} else
		name[0]=0;

	if (type == FLOAT_VECT || type == INT_VECT) {
		nextToken();
		if (sscanf(buffer, "%d", &vlen)==0) {
			error("loader: can't read vector dimension !");
			vlen = 0;
		}
	}

	if (type== OBJECT) {
		// eat the {
		nextToken();
	}

	return true;
}	

bool Loader::checkType(int type) {
	if (type != this->type) {
		fprintf(stderr, "found %s while expecting %s.\n",
				typeNames[this->type],
				typeNames[type]);
		error("check type error");
		return false;
	}
	return true;
}

bool Loader::checkName(const char *name) {

	if (name == 0) 
		return true;

	if (strncmp(name, this->name, 127) != 0) {
		fprintf(stderr, "found %s while expecting %s.\n",
				this->name,
				name);
		error("check name error");
		return false;
	}
	return true;
}

void Loader::getString(char *dest, int maxSize, const char *name) {

	checkType(STRING);
	checkName(name);

	nextToken();
	strncpy(dest, buffer, maxSize-1);
	dest[maxSize-1]=0;
}

float Loader::getFloat(const char *name) {
	float f;

	checkType(FLOAT);
	checkName(name);

	nextToken();
	if (sscanf(buffer, "%f", &f)==0) {
		error("can't read float");
		return 0.0;
	}
	return f;
}

int Loader::getInt(const char *name) {
	int i;

	checkType(INT);
	checkName(name);

	nextToken();
	if (sscanf(buffer, "%d", &i)==0) {
		error("can't read int");
		return 0;
	}
	return i;
}

int Loader::getVectLength() {
	if (type == FLOAT_VECT || type == INT_VECT) {
		return vlen;
	} else {
		error("Trying to get dimension on something which is not vector.");
		return 0;
	}
}

void Loader::getFloatVect(float *vect, const char *name) {

	checkType(FLOAT_VECT);
	checkName(name);

	for (int i=0; i<vlen; i++) {
		nextToken();
		if (sscanf(buffer, "%f", vect + i)==0) {
			error("Can't read float array");
			vect[i]=0;
		}
	}
}

void Loader::getIntVect(int *vect, const char *name) {

	checkType(INT_VECT);
	checkName(name);

	for (int i=0; i<vlen; i++) {
		nextToken();
		if (sscanf(buffer, "%d", vect + i)==0) {
			error("Can't read int array");
			vect[i]=0;
		}
	}
}


