/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Float option for demoNiac
 * leto@calodox.org, 2001
 */

#include <fx.h>
#include "fxinstance.h"
#include "optfloat.h"
#include "interpolator.h"
#include "glmain.h"
#include "report.h"

/*
 * fx.h interface:
 * Register a floating point option.
 * optName :    name of this option
 * flags :      see enums before
 * value :      a pointer to the float, initialized.
 * min, max:    (*value) will be bounded to [min, max]
 */
void registerFloatOpt(char *optName, int flags, float *value,
        float min, float max) {

	if (currentFX)
		currentFX->opts.push_back( (GenericOption *)
			new OptFloat(currentFX, optName, flags, value, min, max) );
}


OptFloat::OptFloat(FXInstance *fxi, char *optName, int flags, float *value,
                float min, float max ) :
		GenericOption(fxi, optName, "Float value", flags) {

	this->fvalue = value; 
	this->value.set(value);

	this->min = min;
	this->max = max;

	if (flags & OPTION_REINIT) {
		track = 0;
	} else {
		track = new TrackFloatVect(1, &this->value, &fixedInterpolator);
	}
}

gint cbOptFloat(GtkWidget *widget, gpointer data) {

	OptFloat *obj = (OptFloat *) data;

	if (obj->ignore) return TRUE;

	debug("cb float");

	*obj->fvalue = gtk_spin_button_get_value_as_float(
		GTK_SPIN_BUTTON(obj->spin)); 

	obj->value.set(obj->fvalue);

	if (obj->track) {
		obj->track->setValue(glMainptr->getTimeMilliSec(), 
				&obj->value);
	}

	glMainptr->reDraw();

	return TRUE;
}

void OptFloat::computeValue(int time) {
	if (track) {
		track->computeValue(time);
		*fvalue = value.asFloat();
	}
}


void OptFloat::draw(GtkBox *box) {
	ignore = false; 

	GtkBox *hbox = drawLabel(box);
	
	// spin button
	GtkAdjustment *adj = (GtkAdjustment *) gtk_adjustment_new(
		value.asFloat(), min, max, 0.1, 10, 1);
	spin = gtk_spin_button_new(adj, 0.1, 3);
	gtk_box_pack_start(hbox, spin, TRUE, TRUE, 5);
	gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spin), flags & OPTION_LIMIT_WRAP);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(spin), TRUE);

	if ((flags & OPTION_REINIT) == 0) {
		gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
			GTK_SIGNAL_FUNC(cbOptFloat), this);
		drawInterp(hbox);
	}
} 

void OptFloat::update() {
	*fvalue = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(spin));
}

void OptFloat::updateGui() {
	ignore = true;
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), *fvalue);
	ignore = false;
}

void OptFloat::save(Saver *s) {
    s->saveObjectBegin("OptFloat");

    s->saveString("name", name);
    s->saveInt("flags", flags);

    if (flags & OPTION_REINIT) {
	s->saveFloat("fvalue", *fvalue);
    } else {
	track->save(s);
    }

    s->saveObjectEnd();
}

void OptFloat::load(Loader *l) {
	assert(strcmp(l->nextName(), "OptFloat")==0);
	l->nextLine();

	name = new char[128];
	l->getString(name, 127);
	l->nextLine();

	flags = l->getInt();
	l->nextLine();

	if (strcmp("fvalue", l->nextName())==0) {
		track = 0;
		*fvalue = l->getFloat();
		l->nextLine();
	} 

	if (l->nextType() == Loader::OBJECT) {
		track = new TrackFloatVect(1, &this->value, &fixedInterpolator);
		track->load(l);
	}

	assert(l->nextType() == Loader::OBJECT_END);
	l->nextLine();
}


