/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * OptVal: class to contain & manage a value of
 * any type for options.
 * leto@calodox.org, 8.2001
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "optval.h"

#ifndef NDEBUG
	#define fatalError(a) fprintf(stderr, "%s(%d): %s\n", __FILE__, __LINE__, a); exit(-1);
#else
	#define fatalError(a) 
#endif 

OptVal::OptVal() {
	clear();
}

OptVal::~OptVal() {
	clear();
}

void OptVal::clear() {
	if (type == STRING && val.s) {
		delete[] val.s;
	}
	type = NONE;
	size=0;
}

OptVal & OptVal::operator = (OptVal &src) {

	if (type == STRING && val.s) {
		// if the dest string is the same
		// than src, string op is useless,
		// and we're done.
		if (src.type == STRING && strcmp(val.s, src.val.s)==0) {
			return *this;
		}

		// free string buffer
		delete[] val.s;
		val.s=0;
	}

	if (src.type == STRING) {
		set(src.asString());
	} else {
		type = src.type;
		val = src.val;
		size = src.size;
	}
	return *this;
}
	
void OptVal::set(int *i, int dim) {
	clear();

	assert(i!=0);

	type = INT;
	size = dim;

	for (int j=0; j<dim; j++) 
		val.iv[j] = i[j]; 
}

void OptVal::set(float *f, int dim) {
	clear();

	assert(f!=0);

	type = FLOAT;
	size = dim;

	for (int j=0; j<dim; j++) 
		val.fv[j] = f[j]; 
}


void OptVal::set(bool b) {
	clear();

	type = BOOL;
	size=1;
	val.b = b;
}

void OptVal::set(char *s) {
	clear();

	type = STRING;
	size=1;
	
	if (s) {
		val.s = new char[strlen(s)+1];
		strcpy(val.s, s);
	} else {
		val.s = 0;
	}
}

int OptVal::asInt() {

	assert(size == 1);

	switch (type) {
		case INT: return val.iv[0];
		case FLOAT: return (int) val.fv[0];
		case BOOL: return (int) val.b;
		default:
			fatalError( "cannot convert value to int.");
	}
	return 0;
}

void OptVal::getInt(int *dest, int dim) {

	assert(dest && dim == size);

	if (type==INT) {
		for (int i=0; i<dim; i++) 
			dest[i] = val.iv[i];
		return;
	}

	if (type == FLOAT) {
		for (int i=0; i<dim; i++) 
			dest[i] = (int) val.fv[i];
		return;
	}

	if (type == BOOL) {
		dest[0] = (int) val.b;
		return;
	}
	fatalError( "cannot convert value to int.");
}

float OptVal::asFloat() {

	assert(size == 1);

	switch (type) {
		case FLOAT: return val.fv[0];
		case INT: return (float) val.iv[0];
		default:
			fatalError( "cannot convert value to float.");
	}
	return 0;
}

void OptVal::getFloat(float *dest, int dim) {

	assert(dest && dim == size);

	if (type == FLOAT) {
		for (int i=0; i<dim; i++) 
			dest[i] = val.fv[i];
		return;
	}

	if (type==INT) {
		for (int i=0; i<dim; i++) 
			dest[i] = (float) val.iv[i];
		return;
	}

	fatalError( "cannot convert value to float.");
}

bool OptVal::asBool() {
	assert(size == 1);

	if (type == BOOL) {
		return val.b;
	}

	if (type == INT) {
		return val.iv[0] != 0;
	}

	fatalError("can't convert to bool !");
	return false;
}

char *OptVal::asString() {

	assert(type == STRING);

	return val.s;
}

void OptVal::save(Saver *s) {
	switch (type) {
		case NONE: break;
		case INT: s->saveIntVect("ival", size, val.iv); break;
		case FLOAT: s->saveFloatVect("fval", size, val.fv); break;
		case BOOL: s->saveInt("bval", (int) val.b); break;
		case STRING: s->saveString("sval", val.s); break;
	}
}

void OptVal::load(Loader *l) {


	if (strcmp("ival", l->nextName()) == 0) {
		int array[OPTVAL_MAX_INT];
		l->getIntVect(array);
		set(array, l->getVectLength());
	} else 
	if (strcmp("fval", l->nextName())==0) {
		float array[OPTVAL_MAX_FLOAT];
		l->getFloatVect(array);
		set(array, l->getVectLength());
	} else 
	if (strcmp("bval", l->nextName()) == 0) {
		set((bool) l->getInt());
	} else 
	if (strcmp("sval", l->nextName()) == 0) {
		char tmp[256];
		l->getString(tmp, 256);
		tmp[255] = 0;
		set(tmp);
	} else {
		l->error("can't load OptVal (unknown type).");
	}

	l->nextLine();
}





