/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Track: keyframing tools for DemoNiac
 * Leto / calodox, 3.2001
 */

#include <stdio.h>
#include <string.h>
#include <math.h> // for abs
#include <stdlib.h> 
#include <assert.h>

#include <matvect.h>

#include "saver.h"
#include "loader.h"
#include "glmain.h"
#include "trackdlg.h"
#include "track.h"
#include "interpolator.h"


#ifdef WIN32
#define MAX_DIM 4
#else
#define MAX_DIM dim
#endif

void Key::save(Saver *s) {
    s->saveInt("key", time);
    value.save(s);
}

void Key::load(Loader *l) {
	time = l->getInt("key");
	l->nextLine();
	value.load(l);
}


/*
 * Generic track
 */

Track::Track( OptVal *value, Interpolator *interpolator) {
	this->variable = value;
	this->interpolator = interpolator;
	first = 0;
	treeItem = 0;
	timeline = 0;
}

void Track::setValue(int time, OptVal *value) {
	interpolator->setValue(time, value, this);
}

void Track::addKey(int time, OptVal *value) {
	Key *newKey = new Key(time, value, this);
	addKey(newKey);

}

void Track::addKeyToList(Key * key) {
	Key *k;

	//printf("New key at %d\n", time);

	for (k=first; k!=0 && k->next !=0; k=k->next) {
		if ( k->next->time >= key->time ) break;
	}

	if (k==0 || first->time >= key->time ) {
		// first element of the list
		key->next = first;
		first = key;
		key->prev = 0;
	} 
	else
	if (k->time < key->time) {
		// let's insert key right after k
		key->prev = k;
		key->next = k->next;
		k->next = key;
	} 

}

void Track::addKey(Key *key) {

	addKeyToList(key);

	// reflect change on GUI
	TrackDlg::addKey(this, key);
}

bool Track::removeKey(int time) {
	Key *k;

	k = getKeyAt(time);
	if (k) {
		// we found the key. Remove it from list.
		removeKey(k);
		return true;
	}
	return false;
}

void Track::removeKey(Key *k) {

	if (k) {

		removeKeyFromList(k);

		// reflect change on GUI
		TrackDlg::removeKey(this, k);

		delete k;
	}
}

void Track::moveKey(Key* key, int time) {

	removeKeyFromList(key);
	key->time = time;
	addKeyToList(key);

	// reflect change to the GUI
	TrackDlg::moveKey(key);
}

void Track::removeKeyFromList(Key *k) {
	if (k->prev)
		k->prev->next = k->next;
	else
		first = k->next;	

	if (k->next)
		k->next->prev = k->prev;
}
	
Key *Track::getKeyAt(int time) {
	Key *k;

	for (k=first; k!=0; k=k->next) {
		if (abs(k->time - time) < GET_KEY_PRECISION) break;
	}
	return k;
}

Key *Track::getKeyBefore(int time) {
	Key *k;

	for (k=first; k!=0 && k->next !=0; k=k->next) {
		if (k->next->time > time + 1) return k;
	}

	return k;
}

void Track::setInterpolator(Interpolator *interpolator) {
	this->interpolator = interpolator;
}
	

void Track::computeValue(int time) {
	Key *k = getKeyBefore(time);

	if (k) {
		if (k->time < time) {
			if (k->next)
				interpolator->getValue(time, k, this);
			else // after the last key.
				*variable = k->value;
		} else {
			// before the first key
			*variable = k->value;
		}
	} 

	// if we need conversion between interpolated
	// value and wanted value.
	//convertValue();
}

bool Track::canInterpolate() {
	return false;
}

void Track::save(Saver *s) {
    s->saveObjectBegin("Track");

    s->saveInt("interpolator", interpolator->getType());

    // dump keys
    for (Key *k=first; k !=0; k=k->next)
	k->save(s);

    s->saveObjectEnd();
}

void Track::load(Loader *l) {

	l->checkName("Track");
	l->nextLine();

	setInterpolator(getInterpByType(l->getInt("interpolator")));
	l->nextLine();

	// get keys
	while (strcmp("key", l->nextName())==0) {
		Key *k = new Key(this);
		k->load(l);
		addKey(k);
	}

	l->checkType(Loader::OBJECT_END);
	l->nextLine();
}
		

/*
 * TrackFloatVect: n-dimentional float vector.
 */

TrackFloatVect::TrackFloatVect(int dim,  OptVal *variable, Interpolator *interpolator)
	: Track(variable, interpolator) {

	this->dim = dim;
}

void TrackFloatVect::interpolateLinear(float t, Key *key) {
	//printf("linear interpolation float between keys: %f\%: %d, %d\n",
	//	t*100.0, key->time, key->next->time);
	float k[MAX_DIM];
	float knext[MAX_DIM];
	float v[MAX_DIM];

	key->value.getFloat(k, dim);
	key->next->value.getFloat(knext, dim);

	for (int i=0; i<dim; i++) {
		v[i] = k[i] + t*(knext[i] - k[i]);
	}


	variable->set(v, dim);
}

void TrackFloatVect::interpolateSmooth(float t, Key *key) {
	double t2 = t*t;
	double t3 = t2*t;

	double h1 = 2.0*t3 -3.0*t2 + 1;
	double h2 = -2*t3 + 3*t2;
	double h3 = t3 -2*t2 + t;
	double h4 = t3-t2;

	float kp[MAX_DIM];
	float k[MAX_DIM];
	float kn[MAX_DIM];
	float knn[MAX_DIM];
	float dest[MAX_DIM];

	if (key->prev) 
		key->prev->value.getFloat(kp, dim);
	else
		key->value.getFloat(kp, dim);
	
	key->value.getFloat(k, dim);
	key->next->value.getFloat(kn, dim);

	if (key->next->next) {
		key->next->next->value.getFloat(knn, dim);
	} else {
		key->next->value.getFloat(knn, dim);
	}	

	for (int i=0; i<dim; i++) {
		// compute tg
		float t1, t2;

		t1 = kn[i] - kp[i];

		t2 = - k[i] + knn[i];

		dest[i] = 
			h1*k[i] +
			h3*t1 + 
			h4*t2 + 
			h2*kn[i]; 
	}
	variable->set(dest, dim);
}

bool TrackFloatVect::canInterpolate() {
	return true;
}

void TrackFloatVect::save(Saver *s) {
    s->saveObjectBegin("TrackFloatVect");

    s->saveInt("interpolator", interpolator->getType());
    s->saveInt("dim", dim);

    // dump keys
    for (Key *k=first; k !=0; k=k->next)
	k->save(s);

    s->saveObjectEnd();
}

void TrackFloatVect::load(Loader *l) {

	assert(	strcmp("TrackFloatVect", l->nextName())==0
		|| strcmp("TrackIntVect", l->nextName())==0);
	l->nextLine();

	setInterpolator(getInterpByType(l->getInt("interpolator")));
	l->nextLine();

	dim = l->getInt("dim");
	l->nextLine();

	// get keys
	while (strcmp("key", l->nextName())==0) {
		Key *k = new Key(this);
		k->load(l);
		addKey(k);
	}

	l->checkType(Loader::OBJECT_END);
	l->nextLine();
}



/*
 * TrackIntVect: n-dimentional int vector, based on TrackFloatVect
 */

TrackIntVect::TrackIntVect(int dim, OptVal *value, Interpolator *interpolator) :
	TrackFloatVect(dim, value, interpolator) 
{
}

void TrackIntVect::setValue(int time, OptVal *value) {
	OptVal v;
	
	v = *value;
	
	TrackFloatVect::setValue(time, value);
	convertValue();
}

void TrackIntVect::convertValue() {
}

void TrackIntVect::save(Saver *s) {
    s->saveObjectBegin("TrackIntVect");

    s->saveInt("interpolator", interpolator->getType());
    s->saveInt("dim", dim);

    // dump keys
    for (Key *k=first; k !=0; k=k->next)
	k->save(s);

    s->saveObjectEnd();
}


/*************************
 *   3D Rotation track   *
 ************************/

Track3DRot::Track3DRot(OptVal *value, Interpolator *interpolator) : Track(value, interpolator) {
}


void Track3DRot::interpolateLinear(float t, Key *key) {
	Quatern q1, q2, res;
	key->value.getFloat((float *)&q1, 4);
	key->next->value.getFloat((float *)&q2, 4);
	QSlerp(&q1, &q2, &res, t);
	variable->set((float *)&res, 4);
}

void Track3DRot::interpolateSmooth(float t, Key *key) {
	interpolateLinear(t, key);
}

void Track3DRot::save(Saver *s) {
    s->saveObjectBegin("Track3DRot");

    s->saveInt("interpolator", interpolator->getType());

    // dump keys
    for (Key *k=first; k !=0; k=k->next)
	k->save(s);

    s->saveObjectEnd();
}

void Track3DRot::load(Loader *l) {

	l->checkName("Track3DRot");
	l->nextLine();

	setInterpolator(getInterpByType(l->getInt("interpolator")));
	l->nextLine();

	// get keys
	while (strcmp("key", l->nextName())==0) {
		Key *k = new Key(this);
		k->load(l);
		addKey(k);
	}

	l->checkType(Loader::OBJECT_END);
	l->nextLine();
}
