/*
 * Calodox Demoniac
 * Copyright (C) 2001 Julien Pilet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * keyframing structure from NCT (Stephane Magnenat) & Leto / calodox
 */

#ifndef _TRACK_H_
#define _TRACK_H_

// that's 1/100 sec.
#define GET_KEY_PRECISION 10

// gtk + gtktimeline is used to talk to TrackDlg
#include <gtk/gtk.h>
#include "gtktimeline.h"

#include "saver.h"
#include "loader.h"
#include "optval.h"

class Track;

class Key {
public:
	Key(Track *track) : time(0), owner(track), adj(0) {
	}

	Key(int time, OptVal *val, Track *track) 
		: time(time), value(), owner(track), adj(0) 
	{
		this->value = *val;
	}

	void save(Saver *s); 
	void load(Loader *l);

	int time;
	OptVal value;

	Key *next;
	Key *prev;

	// the track the key lies on
	Track *owner;

	// for GUI callbacks & managment
	GtkAdjustment *adj;
};

class Interpolator;

/*
 * Track: super class for: 
 * 	TrackFloatVect, TrackIntVect, TrackScreenPos and Track3DRot
 *
 * Objects of class Track (not a subclass) will be
 * tracks for: Bool and String, that can't be interpolated.
 */
class Track {
public:
	Track() {}
	Track(OptVal *value, Interpolator *interpolator);

	virtual void setValue(int time, OptVal *value);

	void addKey(int time, OptVal *value);
	void addKey(Key * key);
	bool removeKey(int time);
	void removeKey(Key *k);
	void moveKey(Key* key, int time);
	Key *getKeyAt(int time);
	Key *getKeyBefore(int time);
	void setInterpolator(Interpolator *interpolator);

	void computeValue(int time);
	virtual void convertValue() { }
	
	virtual bool canInterpolate();
	virtual void interpolateLinear(float t, Key *key) { };
	virtual void interpolateSmooth(float t, Key *key) { };

	virtual void save(Saver *s);
	virtual void load(Loader *l);

	OptVal *variable;
	Interpolator *interpolator;

	// tree node, used by TrackDlg.
	GtkWidget *treeItem;

	// timeline, if present. Used by TrackDlg mainly.
	GtkWidget *timeline;


protected:	

	friend class TrackDlg;

	Key *first;

	void addKeyToList(Key * key);
	void removeKeyFromList(Key * key);
};


/*
 * TrackFloatVect: n-dimentional float vector.
 */

class TrackFloatVect: public Track {
protected:
	int dim;

public:
	TrackFloatVect(int dim, OptVal *variable, Interpolator *interpolator);

	virtual bool canInterpolate();
	virtual void interpolateLinear(float t, Key *key);
	virtual void interpolateSmooth(float t, Key *key);
	virtual void save(Saver *s);
	virtual void load(Loader *l);
};


/*
 * TrackIntVect: n-dimentional int vector, based on TrackFloatVect
 */

class TrackIntVect : public TrackFloatVect {
public:
	TrackIntVect(int dim, OptVal *value, Interpolator *interpolator);

	virtual void setValue(int time, OptVal *value);
	// convert value from float to int
	virtual void convertValue();
	virtual void save(Saver *s);
};


/*
 * Screen position: a 2D int vector clipped to screen coord.
 */

class TrackScreenPos: public TrackFloatVect {
public:
	TrackScreenPos(OptVal *value, Interpolator *interpolator);

	// clip values
	void computeValue(int time);
	virtual void save(Saver *s);
};

// 3D Rotation track.
// I'll code that later.
class Track3DRot : public Track {
public:
	Track3DRot(OptVal *value, Interpolator *interpolator);	
	virtual bool canInterpolate() { return true; }
	virtual void interpolateLinear(float t, Key *key);
	virtual void interpolateSmooth(float t, Key *key);
	virtual void save(Saver *s);
	virtual void load(Loader *l);
};

#endif

